@@ -1,19 +1,44 @@
+use crate::config::{ParseOptions, ParsingMode};
 use crate::id3::v2::header::Id3v2Header;
-use crate::id3::v2::items::Popularimeter;
+use crate::id3::v2::items::PopularimeterFrame;
 use crate::id3::v2::util::pairs::DEFAULT_NUMBER_IN_PAIR;
+use crate::id3::v2::{
+	ChannelInformation, ChannelType, RelativeVolumeAdjustmentFrame, TimestampFrame,
+};
 use crate::picture::MimeType;
-use crate::probe::ParsingMode;
+use crate::tag::items::{Timestamp, ENGLISH};
 use crate::tag::utils::test_utils::read_path;
 
 use super::*;
 
+use std::collections::HashMap;
+
+const COMMENT_FRAME_ID: &str = "COMM";
+
 fn read_tag(path: &str) -> Id3v2Tag {
 	let tag_bytes = read_path(path);
+	read_tag_with_options(
+		&tag_bytes,
+		ParseOptions::new().parsing_mode(ParsingMode::Strict),
+	)
+}
 
-	let mut reader = Cursor::new(&tag_bytes[..]);
+fn read_tag_with_options(bytes: &[u8], parse_options: ParseOptions) -> Id3v2Tag {
+	let mut reader = Cursor::new(bytes);
 
 	let header = Id3v2Header::parse(&mut reader).unwrap();
-	crate::id3::v2::read::parse_id3v2(&mut reader, header, ParsingMode::Strict).unwrap()
+	crate::id3::v2::read::parse_id3v2(&mut reader, header, parse_options).unwrap()
+}
+
+fn dump_and_re_read(tag: &Id3v2Tag, write_options: WriteOptions) -> Id3v2Tag {
+	let mut tag_bytes = Vec::new();
+	let mut writer = Cursor::new(&mut tag_bytes);
+	tag.dump_to(&mut writer, write_options).unwrap();
+
+	read_tag_with_options(
+		&tag_bytes[..],
+		ParseOptions::new().parsing_mode(ParsingMode::Strict),
+	)
 }
 
 #[test]
@@ -21,93 +46,52 @@ fn parse_id3v2() {
 	let mut expected_tag = Id3v2Tag::default();
 
 	let encoding = TextEncoding::Latin1;
-	let flags = FrameFlags::default();
 
-	expected_tag.insert(
-		Frame::new(
-			"TPE1",
-			FrameValue::Text(TextInformationFrame {
-				encoding,
-				value: String::from("Bar artist"),
-			}),
-			flags,
-		)
-		.unwrap(),
-	);
+	expected_tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TPE1")),
+		encoding,
+		String::from("Bar artist"),
+	)));
 
-	expected_tag.insert(
-		Frame::new(
-			"TIT2",
-			FrameValue::Text(TextInformationFrame {
-				encoding,
-				value: String::from("Foo title"),
-			}),
-			flags,
-		)
-		.unwrap(),
-	);
+	expected_tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TIT2")),
+		encoding,
+		String::from("Foo title"),
+	)));
 
-	expected_tag.insert(
-		Frame::new(
-			"TALB",
-			FrameValue::Text(TextInformationFrame {
-				encoding,
-				value: String::from("Baz album"),
-			}),
-			flags,
-		)
-		.unwrap(),
-	);
+	expected_tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TALB")),
+		encoding,
+		String::from("Baz album"),
+	)));
 
-	expected_tag.insert(
-		Frame::new(
-			COMMENT_FRAME_ID,
-			FrameValue::Comment(CommentFrame {
-				encoding,
-				language: *b"eng",
-				description: EMPTY_CONTENT_DESCRIPTOR,
-				content: String::from("Qux comment"),
-			}),
-			flags,
-		)
-		.unwrap(),
-	);
+	expected_tag.insert(Frame::Comment(CommentFrame::new(
+		encoding,
+		*b"eng",
+		EMPTY_CONTENT_DESCRIPTOR,
+		String::from("Qux comment"),
+	)));
 
-	expected_tag.insert(
-		Frame::new(
-			"TDRC",
-			FrameValue::Text(TextInformationFrame {
-				encoding,
-				value: String::from("1984"),
-			}),
-			flags,
-		)
-		.unwrap(),
-	);
+	expected_tag.insert(Frame::Timestamp(TimestampFrame::new(
+		FrameId::Valid(Cow::Borrowed("TDRC")),
+		encoding,
+		Timestamp {
+			year: 1984,
+			..Timestamp::default()
+		},
+	)));
 
-	expected_tag.insert(
-		Frame::new(
-			"TRCK",
-			FrameValue::Text(TextInformationFrame {
-				encoding,
-				value: String::from("1"),
-			}),
-			flags,
-		)
-		.unwrap(),
-	);
+	expected_tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TRCK")),
+		encoding,
+		String::from("1"),
+	)));
 
-	expected_tag.insert(
-		Frame::new(
-			"TCON",
-			FrameValue::Text(TextInformationFrame {
-				encoding,
-				value: String::from("Classical"),
-			}),
-			flags,
-		)
-		.unwrap(),
-	);
+	expected_tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TCON")),
+		encoding,
+		String::from("Classical"),
+	)));
 
 	let parsed_tag = read_tag("tests/tags/assets/id3v2/test.id3v24");
 
@@ -119,13 +103,19 @@ fn id3v2_re_read() {
 	let parsed_tag = read_tag("tests/tags/assets/id3v2/test.id3v24");
 
 	let mut writer = Vec::new();
-	parsed_tag.dump_to(&mut writer).unwrap();
+	parsed_tag
+		.dump_to(&mut writer, WriteOptions::default())
+		.unwrap();
 
 	let temp_reader = &mut &*writer;
 
 	let temp_header = Id3v2Header::parse(temp_reader).unwrap();
-	let temp_parsed_tag =
-		crate::id3::v2::read::parse_id3v2(temp_reader, temp_header, ParsingMode::Strict).unwrap();
+	let temp_parsed_tag = crate::id3::v2::read::parse_id3v2(
+		temp_reader,
+		temp_header,
+		ParseOptions::new().parsing_mode(ParsingMode::Strict),
+	)
+	.unwrap();
 
 	assert_eq!(parsed_tag, temp_parsed_tag);
 }
@@ -139,68 +129,16 @@ fn id3v2_to_tag() {
 	crate::tag::utils::test_utils::verify_tag(&tag, true, true);
 }
 
-#[test]
-fn id3v2_to_tag_popm() {
-	let id3v2 = read_tag("tests/tags/assets/id3v2/test_popm.id3v24");
-
-	let tag: Tag = id3v2.into();
-
-	assert_eq!(
-		tag.get_binary(&ItemKey::Popularimeter, false),
-		Some(
-			&[
-				b'f', b'o', b'o', b'@', b'b', b'a', b'r', b'.', b'c', b'o', b'm', 0, 196, 0, 0,
-				255, 255,
-			][..]
-		),
-	);
-}
-
-#[test]
-fn tag_to_id3v2_popm() {
-	let mut tag = Tag::new(TagType::Id3v2);
-	tag.insert(TagItem::new(
-		ItemKey::Popularimeter,
-		ItemValue::Binary(vec![
-			b'f', b'o', b'o', b'@', b'b', b'a', b'r', b'.', b'c', b'o', b'm', 0, 196, 0, 0, 255,
-			255,
-		]),
-	));
-
-	let expected = Popularimeter {
-		email: String::from("foo@bar.com"),
-		rating: 196,
-		counter: 65535,
-	};
-
-	let converted_tag: Id3v2Tag = tag.into();
-
-	assert_eq!(converted_tag.frames.len(), 1);
-	let actual_frame = converted_tag.frames.first().unwrap();
-
-	assert_eq!(actual_frame.id, FrameId::Valid(Cow::Borrowed("POPM")));
-	// Note: as POPM frames are considered equal by email alone, each field must
-	// be separately validated
-	match actual_frame.content() {
-		FrameValue::Popularimeter(pop) => {
-			assert_eq!(pop.email, expected.email);
-			assert_eq!(pop.rating, expected.rating);
-			assert_eq!(pop.counter, expected.counter);
-		},
-		_ => unreachable!(),
-	}
-}
-
 #[test]
 fn fail_write_bad_frame() {
 	let mut tag = Id3v2Tag::default();
-	tag.insert(Frame {
-		id: FrameId::Valid(Cow::Borrowed("ABCD")),
-		value: FrameValue::Url(UrlLinkFrame(String::from("FOO URL"))),
-		flags: FrameFlags::default(),
-	});
 
-	let res = tag.dump_to(&mut Vec::<u8>::new());
+	tag.insert(Frame::Url(UrlLinkFrame::new(
+		FrameId::Valid(Cow::Borrowed("ABCD")),
+		String::from("FOO URL"),
+	)));
+
+	let res = tag.dump_to(&mut Vec::<u8>::new(), WriteOptions::default());
 
 	assert!(res.is_err());
 	assert_eq!(
@@ -219,11 +157,12 @@ fn tag_to_id3v2() {
 		let frame = frame.unwrap();
 
 		assert_eq!(
-			frame.content(),
-			&FrameValue::Text(TextInformationFrame {
-				encoding: TextEncoding::UTF8,
-				value: String::from(value)
-			})
+			frame,
+			&Frame::Text(TextInformationFrame::new(
+				FrameId::Valid(Cow::Borrowed(id)),
+				TextEncoding::UTF8,
+				String::from(value)
+			)),
 		);
 	}
 
@@ -239,13 +178,13 @@ fn tag_to_id3v2() {
 		.get(&FrameId::Valid(Cow::Borrowed(COMMENT_FRAME_ID)))
 		.unwrap();
 	assert_eq!(
-		frame.content(),
-		&FrameValue::Comment(CommentFrame {
-			encoding: TextEncoding::Latin1,
-			language: *b"eng",
-			description: EMPTY_CONTENT_DESCRIPTOR,
-			content: String::from("Qux comment")
-		})
+		frame,
+		&Frame::Comment(CommentFrame::new(
+			TextEncoding::UTF8,
+			*b"XXX",
+			EMPTY_CONTENT_DESCRIPTOR,
+			String::from("Qux comment"),
+		))
 	);
 
 	verify_frame(&id3v2_tag, "TRCK", "1");
@@ -258,84 +197,61 @@ fn create_full_test_tag(version: Id3v2Version) -> Id3v2Tag {
 	tag.original_version = version;
 
 	let encoding = TextEncoding::UTF16;
-	let flags = FrameFlags::default();
 
-	tag.insert(Frame {
-		id: FrameId::Valid(Cow::Borrowed("TIT2")),
-		value: FrameValue::Text(TextInformationFrame {
-			encoding,
-			value: String::from("TempleOS Hymn Risen (Remix)"),
-		}),
-		flags,
-	});
+	tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TIT2")),
+		encoding,
+		String::from("TempleOS Hymn Risen (Remix)"),
+	)));
 
-	tag.insert(Frame {
-		id: FrameId::Valid(Cow::Borrowed("TPE1")),
-		value: FrameValue::Text(TextInformationFrame {
-			encoding,
-			value: String::from("Dave Eddy"),
-		}),
-		flags,
-	});
+	tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TPE1")),
+		encoding,
+		String::from("Dave Eddy"),
+	)));
 
-	tag.insert(Frame {
-		id: FrameId::Valid(Cow::Borrowed("TRCK")),
-		value: FrameValue::Text(TextInformationFrame {
-			encoding: TextEncoding::Latin1,
-			value: String::from("1"),
-		}),
-		flags,
-	});
+	tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TRCK")),
+		encoding,
+		String::from("1"),
+	)));
 
-	tag.insert(Frame {
-		id: FrameId::Valid(Cow::Borrowed("TALB")),
-		value: FrameValue::Text(TextInformationFrame {
-			encoding,
-			value: String::from("Summer"),
-		}),
-		flags,
-	});
+	tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TALB")),
+		encoding,
+		String::from("Summer"),
+	)));
 
-	tag.insert(Frame {
-		id: FrameId::Valid(Cow::Borrowed("TDRC")),
-		value: FrameValue::Text(TextInformationFrame {
-			encoding,
-			value: String::from("2017"),
-		}),
-		flags,
-	});
+	tag.insert(Frame::Timestamp(TimestampFrame::new(
+		FrameId::Valid(Cow::Borrowed("TDRC")),
+		encoding,
+		Timestamp {
+			year: 2017,
+			..Timestamp::default()
+		},
+	)));
 
-	tag.insert(Frame {
-		id: FrameId::Valid(Cow::Borrowed("TCON")),
-		value: FrameValue::Text(TextInformationFrame {
-			encoding,
-			value: String::from("Electronic"),
-		}),
-		flags,
-	});
+	tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TCON")),
+		encoding,
+		String::from("Electronic"),
+	)));
 
-	tag.insert(Frame {
-		id: FrameId::Valid(Cow::Borrowed("TLEN")),
-		value: FrameValue::Text(TextInformationFrame {
-			encoding: TextEncoding::UTF16,
-			value: String::from("213017"),
-		}),
-		flags,
-	});
+	tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::Borrowed("TLEN")),
+		encoding,
+		String::from("213017"),
+	)));
 
-	tag.insert(Frame {
-		id: FrameId::Valid(Cow::Borrowed("APIC")),
-		value: FrameValue::Picture(AttachedPictureFrame {
-			encoding: TextEncoding::Latin1,
-			picture: Picture {
-				pic_type: PictureType::CoverFront,
-				mime_type: Some(MimeType::Png),
-				description: None,
-				data: read_path("tests/tags/assets/id3v2/test_full_cover.png").into(),
-			},
-		}),
-		flags,
-	});
+	tag.insert(Frame::Picture(AttachedPictureFrame::new(
+		TextEncoding::Latin1,
+		Picture {
+			pic_type: PictureType::CoverFront,
+			mime_type: Some(MimeType::Png),
+			description: None,
+			data: read_path("tests/tags/assets/id3v2/test_full_cover.png").into(),
+		},
+	)));
 
 	tag
 }
@@ -350,9 +266,14 @@ fn id3v24_full() {
 
 #[test]
 fn id3v23_full() {
-	let tag = create_full_test_tag(Id3v2Version::V3);
-	let parsed_tag = read_tag("tests/tags/assets/id3v2/test_full.id3v23");
+	let mut tag = create_full_test_tag(Id3v2Version::V3);
+	let mut parsed_tag = read_tag("tests/tags/assets/id3v2/test_full.id3v23");
 
+	// Tags may change order after being read, due to the TDRC conversion
+	tag.frames.sort_by_key(|frame| frame.id_str().to_string());
+	parsed_tag
+		.frames
+		.sort_by_key(|frame| frame.id_str().to_string());
 	assert_eq!(tag, parsed_tag);
 }
 
@@ -370,12 +291,17 @@ fn id3v24_footer() {
 	tag.flags.footer = true;
 
 	let mut writer = Vec::new();
-	tag.dump_to(&mut writer).unwrap();
+	tag.dump_to(&mut writer, WriteOptions::default()).unwrap();
 
 	let mut reader = &mut &writer[..];
 
 	let header = Id3v2Header::parse(&mut reader).unwrap();
-	assert!(crate::id3::v2::read::parse_id3v2(reader, header, ParsingMode::Strict).is_ok());
+	let _ = crate::id3::v2::read::parse_id3v2(
+		reader,
+		header,
+		ParseOptions::new().parsing_mode(ParsingMode::Strict),
+	)
+	.unwrap();
 
 	assert_eq!(writer[3..10], writer[writer.len() - 7..])
 }
@@ -395,24 +321,25 @@ fn issue_36() {
 	tag.push_picture(picture.clone());
 
 	let mut writer = Vec::new();
-	tag.dump_to(&mut writer).unwrap();
+	tag.dump_to(&mut writer, WriteOptions::default()).unwrap();
 
 	let mut reader = &mut &writer[..];
 
 	let header = Id3v2Header::parse(&mut reader).unwrap();
-	let tag = crate::id3::v2::read::parse_id3v2(reader, header, ParsingMode::Strict).unwrap();
+	let tag = crate::id3::v2::read::parse_id3v2(
+		reader,
+		header,
+		ParseOptions::new().parsing_mode(ParsingMode::Strict),
+	)
+	.unwrap();
 
 	assert_eq!(tag.len(), 1);
 	assert_eq!(
 		tag.frames.first(),
-		Some(&Frame {
-			id: FrameId::Valid(Cow::Borrowed("APIC")),
-			value: FrameValue::Picture(AttachedPictureFrame {
-				encoding: TextEncoding::UTF8,
-				picture
-			}),
-			flags: FrameFlags::default()
-		})
+		Some(&Frame::Picture(AttachedPictureFrame::new(
+			TextEncoding::UTF8,
+			picture
+		)))
 	);
 }
 
@@ -423,20 +350,19 @@ fn popm_frame() {
 	assert_eq!(parsed_tag.frames.len(), 1);
 	let popm_frame = parsed_tag.frames.first().unwrap();
 
-	assert_eq!(popm_frame.id, FrameId::Valid(Cow::Borrowed("POPM")));
+	assert_eq!(popm_frame.id(), &FrameId::Valid(Cow::Borrowed("POPM")));
 	assert_eq!(
-		popm_frame.value,
-		FrameValue::Popularimeter(Popularimeter {
-			email: String::from("foo@bar.com"),
-			rating: 196,
-			counter: 65535
-		})
+		popm_frame,
+		&Frame::Popularimeter(PopularimeterFrame::new(
+			String::from("foo@bar.com"),
+			196,
+			65535
+		))
 	)
 }
 
 #[test]
 fn multi_value_frame_to_tag() {
-	use crate::traits::Accessor;
 	let mut tag = Id3v2Tag::default();
 
 	tag.set_artist(String::from("foo\0bar\0baz"));
@@ -448,7 +374,6 @@ fn multi_value_frame_to_tag() {
 
 #[test]
 fn multi_item_tag_to_id3v2() {
-	use crate::traits::Accessor;
 	let mut tag = Tag::new(TagType::Id3v2);
 
 	tag.push_unchecked(TagItem::new(
@@ -476,28 +401,21 @@ fn utf16_txxx_with_single_bom() {
 #[test]
 fn replaygain_tag_conversion() {
 	let mut tag = Id3v2Tag::default();
-	tag.insert(
-		Frame::new(
-			"TXXX",
-			FrameValue::UserText(ExtendedTextFrame {
-				encoding: TextEncoding::UTF8,
-				description: String::from("REPLAYGAIN_ALBUM_GAIN"),
-				content: String::from("-10.43 dB"),
-			}),
-			FrameFlags::default(),
-		)
-		.unwrap(),
-	);
+	tag.insert(Frame::UserText(ExtendedTextFrame::new(
+		TextEncoding::UTF8,
+		String::from("REPLAYGAIN_ALBUM_GAIN"),
+		String::from("-10.43 dB"),
+	)));
 
 	let tag: Tag = tag.into();
 
 	assert_eq!(tag.item_count(), 1);
 	assert_eq!(
 		tag.items[0],
-		TagItem {
-			item_key: ItemKey::ReplayGainAlbumGain,
-			item_value: ItemValue::Text(String::from("-10.43 dB"))
-		}
+		TagItem::new(
+			ItemKey::ReplayGainAlbumGain,
+			ItemValue::Text(String::from("-10.43 dB"))
+		)
 	);
 }
 
@@ -572,7 +490,6 @@ fn multi_value_roundtrip() {
 fn comments() {
 	let mut tag = Id3v2Tag::default();
 	let encoding = TextEncoding::Latin1;
-	let flags = FrameFlags::default();
 	let custom_descriptor = "lofty-rs";
 
 	assert!(tag.comment().is_none());
@@ -587,19 +504,12 @@ fn comments() {
 		.iter()
 		.find_map(|frame| filter_comment_frame_by_description(frame, custom_descriptor))
 		.is_none());
-	tag.insert(
-		Frame::new(
-			COMMENT_FRAME_ID,
-			FrameValue::Comment(CommentFrame {
-				encoding,
-				language: *b"eng",
-				description: custom_descriptor.to_owned(),
-				content: String::from("Qux comment"),
-			}),
-			flags,
-		)
-		.unwrap(),
-	);
+	tag.insert(Frame::Comment(CommentFrame::new(
+		encoding,
+		*b"eng",
+		custom_descriptor.to_owned(),
+		String::from("Qux comment"),
+	)));
 	// Verify that the regular comment still exists
 	assert_eq!(Some(Cow::Borrowed("")), tag.comment());
 	assert_eq!(1, tag.comments().count());
@@ -617,27 +527,17 @@ fn comments() {
 
 #[test]
 fn txxx_wxxx_tag_conversion() {
-	let txxx_frame = Frame::new(
-		"TXXX",
-		FrameValue::UserText(ExtendedTextFrame {
-			encoding: TextEncoding::UTF8,
-			description: String::from("FOO_TEXT_FRAME"),
-			content: String::from("foo content"),
-		}),
-		FrameFlags::default(),
-	)
-	.unwrap();
+	let txxx_frame = Frame::UserText(ExtendedTextFrame::new(
+		TextEncoding::UTF8,
+		String::from("FOO_TEXT_FRAME"),
+		String::from("foo content"),
+	));
 
-	let wxxx_frame = Frame::new(
-		"WXXX",
-		FrameValue::UserUrl(ExtendedUrlFrame {
-			encoding: TextEncoding::UTF8,
-			description: String::from("BAR_URL_FRAME"),
-			content: String::from("bar url"),
-		}),
-		FrameFlags::default(),
-	)
-	.unwrap();
+	let wxxx_frame = Frame::UserUrl(ExtendedUrlFrame::new(
+		TextEncoding::UTF8,
+		String::from("BAR_URL_FRAME"),
+		String::from("bar url"),
+	));
 
 	let mut tag = Id3v2Tag::default();
 
@@ -671,18 +571,11 @@ fn txxx_wxxx_tag_conversion() {
 #[test]
 fn user_defined_frames_conversion() {
 	let mut id3v2 = Id3v2Tag::default();
-	id3v2.insert(
-		Frame::new(
-			"TXXX",
-			FrameValue::UserText(ExtendedTextFrame {
-				encoding: TextEncoding::UTF8,
-				description: String::from("FOO_BAR"),
-				content: String::from("foo content"),
-			}),
-			FrameFlags::default(),
-		)
-		.unwrap(),
-	);
+	id3v2.insert(Frame::UserText(ExtendedTextFrame::new(
+		TextEncoding::UTF8,
+		String::from("FOO_BAR"),
+		String::from("foo content"),
+	)));
 
 	let (split_remainder, split_tag) = id3v2.split_tag();
 	assert_eq!(split_remainder.0.len(), 0);
@@ -693,15 +586,11 @@ fn user_defined_frames_conversion() {
 	// Verify we properly convert user defined frames between Tag <-> ID3v2Tag round trips
 	assert_eq!(
 		id3v2.frames.first(),
-		Some(&Frame {
-			id: FrameId::Valid(Cow::Borrowed("TXXX")),
-			value: FrameValue::UserText(ExtendedTextFrame {
-				description: String::from("FOO_BAR"),
-				encoding: TextEncoding::UTF8, // Not considered by PartialEq!
-				content: String::new(),       // Not considered by PartialEq!
-			}),
-			flags: FrameFlags::default(),
-		})
+		Some(&Frame::UserText(ExtendedTextFrame::new(
+			TextEncoding::UTF8, // Not considered by PartialEq!
+			String::from("FOO_BAR"),
+			String::new(), // Not considered by PartialEq!
+		),))
 	);
 
 	// Verify we properly convert user defined frames when writing a Tag, which has to convert
@@ -710,15 +599,19 @@ fn user_defined_frames_conversion() {
 	assert_eq!(tag.len(), 1);
 
 	let mut content = Vec::new();
-	tag.dump_to(&mut content).unwrap();
+	tag.dump_to(&mut content, WriteOptions::default()).unwrap();
 	assert!(!content.is_empty());
 
 	// And verify we can reread the tag
 	let mut reader = std::io::Cursor::new(&content[..]);
 
 	let header = Id3v2Header::parse(&mut reader).unwrap();
-	let reparsed =
-		crate::id3::v2::read::parse_id3v2(&mut reader, header, ParsingMode::Strict).unwrap();
+	let reparsed = crate::id3::v2::read::parse_id3v2(
+		&mut reader,
+		header,
+		ParseOptions::new().parsing_mode(ParsingMode::Strict),
+	)
+	.unwrap();
 
 	assert_eq!(id3v2, reparsed);
 }
@@ -821,7 +714,6 @@ fn set_disk_total_and_disk() {
 
 #[test]
 fn track_number_tag_to_id3v2() {
-	use crate::traits::Accessor;
 	let track_number = 1;
 
 	let mut tag = Tag::new(TagType::Id3v2);
@@ -839,7 +731,6 @@ fn track_number_tag_to_id3v2() {
 
 #[test]
 fn track_total_tag_to_id3v2() {
-	use crate::traits::Accessor;
 	let track_total = 2;
 
 	let mut tag = Tag::new(TagType::Id3v2);
@@ -857,7 +748,6 @@ fn track_total_tag_to_id3v2() {
 
 #[test]
 fn track_number_and_track_total_tag_to_id3v2() {
-	use crate::traits::Accessor;
 	let track_number = 1;
 	let track_total = 2;
 
@@ -881,7 +771,6 @@ fn track_number_and_track_total_tag_to_id3v2() {
 
 #[test]
 fn disk_number_tag_to_id3v2() {
-	use crate::traits::Accessor;
 	let disk_number = 1;
 
 	let mut tag = Tag::new(TagType::Id3v2);
@@ -899,7 +788,6 @@ fn disk_number_tag_to_id3v2() {
 
 #[test]
 fn disk_total_tag_to_id3v2() {
-	use crate::traits::Accessor;
 	let disk_total = 2;
 
 	let mut tag = Tag::new(TagType::Id3v2);
@@ -917,7 +805,6 @@ fn disk_total_tag_to_id3v2() {
 
 #[test]
 fn disk_number_and_disk_total_tag_to_id3v2() {
-	use crate::traits::Accessor;
 	let disk_number = 1;
 	let disk_total = 2;
 
@@ -944,7 +831,6 @@ fn create_tag_with_trck_and_tpos_frame(content: &'static str) -> Tag {
 		tag.insert(new_text_frame(
 			FrameId::Valid(Cow::Borrowed(id)),
 			content.to_string(),
-			FrameFlags::default(),
 		));
 	}
 
@@ -1001,31 +887,17 @@ fn invalid_trck_and_tpos_frame() {
 #[test]
 fn ufid_frame_with_musicbrainz_record_id() {
 	let mut id3v2 = Id3v2Tag::default();
-	let unknown_ufid_frame = UniqueFileIdentifierFrame {
-		owner: "other".to_owned(),
-		identifier: b"0123456789".to_vec(),
-	};
-	id3v2.insert(
-		Frame::new(
-			"UFID",
-			FrameValue::UniqueFileIdentifier(unknown_ufid_frame.clone()),
-			FrameFlags::default(),
-		)
-		.unwrap(),
-	);
+	let unknown_ufid_frame =
+		UniqueFileIdentifierFrame::new("other".to_owned(), b"0123456789".to_vec());
+	id3v2.insert(Frame::UniqueFileIdentifier(unknown_ufid_frame.clone()));
 	let musicbrainz_recording_id = b"189002e7-3285-4e2e-92a3-7f6c30d407a2";
-	let musicbrainz_recording_id_frame = UniqueFileIdentifierFrame {
-		owner: MUSICBRAINZ_UFID_OWNER.to_owned(),
-		identifier: musicbrainz_recording_id.to_vec(),
-	};
-	id3v2.insert(
-		Frame::new(
-			"UFID",
-			FrameValue::UniqueFileIdentifier(musicbrainz_recording_id_frame.clone()),
-			FrameFlags::default(),
-		)
-		.unwrap(),
+	let musicbrainz_recording_id_frame = UniqueFileIdentifierFrame::new(
+		MUSICBRAINZ_UFID_OWNER.to_owned(),
+		musicbrainz_recording_id.to_vec(),
 	);
+	id3v2.insert(Frame::UniqueFileIdentifier(
+		musicbrainz_recording_id_frame.clone(),
+	));
 	assert_eq!(2, id3v2.len());
 
 	let (split_remainder, split_tag) = id3v2.split_tag();
@@ -1043,23 +915,15 @@ fn ufid_frame_with_musicbrainz_record_id() {
 	let id3v2 = split_remainder.merge_tag(split_tag);
 	assert_eq!(2, id3v2.len());
 	match &id3v2.frames[..] {
-		[Frame {
-			id: _,
-			value:
-				FrameValue::UniqueFileIdentifier(UniqueFileIdentifierFrame {
-					owner: first_owner,
-					identifier: first_identifier,
-				}),
-			flags: _,
-		}, Frame {
-			id: _,
-			value:
-				FrameValue::UniqueFileIdentifier(UniqueFileIdentifierFrame {
-					owner: second_owner,
-					identifier: second_identifier,
-				}),
-			flags: _,
-		}] => {
+		[Frame::UniqueFileIdentifier(UniqueFileIdentifierFrame {
+			owner: first_owner,
+			identifier: first_identifier,
+			..
+		}), Frame::UniqueFileIdentifier(UniqueFileIdentifierFrame {
+			owner: second_owner,
+			identifier: second_identifier,
+			..
+		})] => {
 			assert_eq!(&unknown_ufid_frame.owner, first_owner);
 			assert_eq!(&unknown_ufid_frame.identifier, first_identifier);
 			assert_eq!(&musicbrainz_recording_id_frame.owner, second_owner);
@@ -1080,30 +944,20 @@ fn get_set_user_defined_text() {
 	let content2 = String::new();
 
 	let mut id3v2 = Id3v2Tag::default();
-	let txxx_frame = Frame::new(
-		"TXXX",
-		ExtendedTextFrame {
-			encoding: TextEncoding::UTF8,
-			description: description.clone(),
-			content: content.clone(),
-		},
-		FrameFlags::default(),
-	)
-	.unwrap();
+	let txxx_frame = Frame::UserText(ExtendedTextFrame::new(
+		TextEncoding::UTF8,
+		description.clone(),
+		content.clone(),
+	));
 
 	id3v2.insert(txxx_frame.clone());
 
 	// Insert another to verify we can search through multiple
-	let txxx_frame2 = Frame::new(
-		"TXXX",
-		ExtendedTextFrame {
-			encoding: TextEncoding::UTF8,
-			description: description2.clone(),
-			content: content2.clone(),
-		},
-		FrameFlags::default(),
-	)
-	.unwrap();
+	let txxx_frame2 = Frame::UserText(ExtendedTextFrame::new(
+		TextEncoding::UTF8,
+		description2.clone(),
+		content2.clone(),
+	));
 	id3v2.insert(txxx_frame2);
 
 	// We cannot get user defined texts through `get_text`
@@ -1167,7 +1021,7 @@ fn trim_end_nulls_when_reading_frame_content() {
 	assert_eq!(comment, "Comment");
 
 	let url_frame = tag.get(&FrameId::Valid(Cow::Borrowed("WXXX"))).unwrap();
-	let FrameValue::UserUrl(url) = &url_frame.value else {
+	let Frame::UserUrl(url) = &url_frame else {
 		panic!("Expected a UserUrl")
 	};
 	assert_eq!(url.content, "https://www.myfanpage.com");
@@ -1175,7 +1029,7 @@ fn trim_end_nulls_when_reading_frame_content() {
 
 fn id3v2_tag_with_genre(value: &str) -> Id3v2Tag {
 	let mut tag = Id3v2Tag::default();
-	let frame = new_text_frame(GENRE_ID, String::from(value), FrameFlags::default());
+	let frame = new_text_frame(GENRE_ID, String::from(value));
 	tag.insert(frame);
 	tag
 }
@@ -1267,10 +1121,11 @@ fn genres_id_remix_cover() {
 #[test]
 fn tipl_round_trip() {
 	let mut tag = Id3v2Tag::default();
-	let mut tipl = KeyValueFrame {
-		encoding: TextEncoding::UTF8,
-		key_value_pairs: Vec::new(),
-	};
+	let mut tipl = KeyValueFrame::new(
+		FrameId::Valid(Cow::Borrowed("TIPL")),
+		TextEncoding::UTF8,
+		Vec::new(),
+	);
 
 	// Add all supported keys
 	for (_, key) in TIPL_MAPPINGS {
@@ -1282,14 +1137,7 @@ fn tipl_round_trip() {
 	tipl.key_value_pairs
 		.push((String::from("Foo"), String::from("Bar")));
 
-	tag.insert(
-		Frame::new(
-			"TIPL",
-			FrameValue::KeyValue(tipl.clone()),
-			FrameFlags::default(),
-		)
-		.unwrap(),
-	);
+	tag.insert(Frame::KeyValue(tipl.clone()));
 
 	let (split_remainder, split_tag) = tag.split_tag();
 	assert_eq!(split_remainder.0.len(), 1); // "Foo" is not supported
@@ -1308,11 +1156,7 @@ fn tipl_round_trip() {
 	let mut id3v2 = split_remainder.merge_tag(split_tag);
 	assert_eq!(id3v2.frames.len(), 1);
 	match &mut id3v2.frames[..] {
-		[Frame {
-			id: _,
-			value: FrameValue::KeyValue(tipl2),
-			flags: _,
-		}] => {
+		[Frame::KeyValue(tipl2)] => {
 			// Order will not be the same, so we have to sort first
 			tipl.key_value_pairs.sort();
 			tipl2.key_value_pairs.sort();
@@ -1340,7 +1184,353 @@ fn flag_item_conversion() {
 }
 
 #[test]
-fn regression_295_fuzzing_id3v24_parse_extended_text_frame_unreachable_panic() {
-	let data = [1, 0, 0, 0];
-	assert!(ExtendedTextFrame::parse(&mut &data[..], Id3v2Version::V4).is_err());
+fn itunes_advisory_roundtrip() {
+	use crate::mp4::{AdvisoryRating, Ilst};
+
+	let mut tag = Ilst::new();
+	tag.set_advisory_rating(AdvisoryRating::Explicit);
+
+	let tag: Tag = tag.into();
+	let tag: Id3v2Tag = tag.into();
+
+	assert_eq!(tag.frames.len(), 1);
+
+	let frame = tag.get_user_text("ITUNESADVISORY");
+	assert!(frame.is_some());
+	assert_eq!(frame.unwrap(), "1");
+
+	let tag: Tag = tag.into();
+	let tag: Ilst = tag.into();
+
+	assert_eq!(tag.advisory_rating(), Some(AdvisoryRating::Explicit));
+}
+
+#[test]
+fn timestamp_roundtrip() {
+	let mut tag = Id3v2Tag::default();
+	tag.insert(Frame::Timestamp(TimestampFrame::new(
+		FrameId::Valid(Cow::Borrowed("TDRC")),
+		TextEncoding::UTF8,
+		Timestamp {
+			year: 2024,
+			month: Some(6),
+			day: Some(3),
+			hour: Some(14),
+			minute: Some(8),
+			second: Some(49),
+		},
+	)));
+
+	let tag: Tag = tag.into();
+	assert_eq!(tag.len(), 1);
+	assert_eq!(
+		tag.get_string(&ItemKey::RecordingDate),
+		Some("2024-06-03T14:08:49")
+	);
+
+	let tag: Id3v2Tag = tag.into();
+	assert_eq!(tag.frames.len(), 1);
+
+	let frame = tag.frames.first().unwrap();
+	assert_eq!(frame.id(), &FrameId::Valid(Cow::Borrowed("TDRC")));
+	match &frame {
+		Frame::Timestamp(frame) => {
+			assert_eq!(frame.timestamp.year, 2024);
+			assert_eq!(frame.timestamp.month, Some(6));
+			assert_eq!(frame.timestamp.day, Some(3));
+			assert_eq!(frame.timestamp.hour, Some(14));
+			assert_eq!(frame.timestamp.minute, Some(8));
+			assert_eq!(frame.timestamp.second, Some(49));
+		},
+		_ => panic!("Expected a TimestampFrame"),
+	}
+}
+
+#[test]
+fn special_items_roundtrip() {
+	let mut tag = Id3v2Tag::new();
+
+	let rva2 = Frame::RelativeVolumeAdjustment(RelativeVolumeAdjustmentFrame::new(
+		String::from("Foo RVA"),
+		HashMap::from([(
+			ChannelType::MasterVolume,
+			ChannelInformation {
+				channel_type: ChannelType::MasterVolume,
+				volume_adjustment: 30,
+				bits_representing_peak: 0,
+				peak_volume: None,
+			},
+		)]),
+	));
+
+	tag.insert(rva2.clone());
+	tag.set_artist(String::from("Foo Artist")); // Some value that we *can* represent generically
+
+	let tag: Tag = tag.into();
+
+	assert_eq!(tag.len(), 1);
+	assert_eq!(tag.artist().as_deref(), Some("Foo Artist"));
+
+	let mut tag: Id3v2Tag = tag.into();
+
+	assert_eq!(tag.frames.len(), 2);
+	assert_eq!(tag.artist().as_deref(), Some("Foo Artist"));
+	assert_eq!(tag.get(&FrameId::Valid(Cow::Borrowed("RVA2"))), Some(&rva2));
+
+	let mut tag_bytes = Vec::new();
+	tag.dump_to(&mut tag_bytes, WriteOptions::default())
+		.unwrap();
+
+	let mut tag_re_read = read_tag_with_options(
+		&tag_bytes[..],
+		ParseOptions::new().parsing_mode(ParsingMode::Strict),
+	);
+
+	// Ensure ordered comparison
+	tag.frames.sort_by_key(|frame| frame.id().to_string());
+	tag_re_read
+		.frames
+		.sort_by_key(|frame| frame.id().to_string());
+	assert_eq!(tag, tag_re_read);
+
+	// Now write from `Tag`
+	let tag: Tag = tag.into();
+
+	let mut tag_bytes = Vec::new();
+	tag.dump_to(&mut tag_bytes, WriteOptions::default())
+		.unwrap();
+
+	let mut generic_tag_re_read = read_tag_with_options(
+		&tag_bytes[..],
+		ParseOptions::new().parsing_mode(ParsingMode::Strict),
+	);
+
+	generic_tag_re_read
+		.frames
+		.sort_by_key(|frame| frame.id().to_string());
+	assert_eq!(tag_re_read, generic_tag_re_read);
+}
+
+#[test]
+fn preserve_comment_lang_description_on_conversion() {
+	let mut tag = Id3v2Tag::new();
+
+	let comment_frame = Frame::Comment(CommentFrame::new(
+		TextEncoding::UTF8,
+		ENGLISH,
+		String::from("Some description"),
+		String::from("Foo comment"),
+	));
+
+	tag.insert(comment_frame.clone());
+
+	let tag: Tag = tag.into();
+	assert_eq!(tag.len(), 1);
+
+	let tag: Id3v2Tag = tag.into();
+	assert_eq!(tag.len(), 1);
+
+	let frame = tag.get(&FrameId::Valid(Cow::Borrowed("COMM"))).unwrap();
+	match frame {
+		Frame::Comment(comm) => {
+			assert_eq!(comm.language, ENGLISH);
+			assert_eq!(comm.description, "Some description");
+			assert_eq!(comm.content, "Foo comment");
+		},
+		_ => panic!("Expected a CommentFrame"),
+	}
+}
+
+// TODO: Remove this once we have a better solution
+#[test]
+fn hold_back_4_character_txxx_description() {
+	let mut tag = Id3v2Tag::new();
+
+	let _ = tag.insert_user_text(String::from("MODE"), String::from("CBR"));
+
+	let tag: Tag = tag.into();
+	assert_eq!(tag.len(), 0);
+
+	let tag: Id3v2Tag = tag.into();
+	assert_eq!(tag.len(), 1);
+}
+
+#[test]
+fn skip_reading_cover_art() {
+	let p = Picture::new_unchecked(
+		PictureType::CoverFront,
+		Some(MimeType::Jpeg),
+		None,
+		std::iter::repeat(0).take(50).collect::<Vec<u8>>(),
+	);
+
+	let mut tag = Tag::new(TagType::Id3v2);
+	tag.push_picture(p);
+
+	tag.set_artist(String::from("Foo artist"));
+
+	let mut writer = Vec::new();
+	tag.dump_to(&mut writer, WriteOptions::new()).unwrap();
+
+	let id3v2 = read_tag_with_options(&writer[..], ParseOptions::new().read_cover_art(false));
+	assert_eq!(id3v2.len(), 1); // Artist, no picture
+	assert!(id3v2.artist().is_some());
+}
+
+#[test]
+fn remove_id3v24_frames_on_id3v23_save() {
+	let mut tag = Id3v2Tag::new();
+
+	tag.insert(Frame::RelativeVolumeAdjustment(
+		RelativeVolumeAdjustmentFrame::new(
+			String::from("Foo RVA"),
+			HashMap::from([(
+				ChannelType::MasterVolume,
+				ChannelInformation {
+					channel_type: ChannelType::MasterVolume,
+					volume_adjustment: 30,
+					bits_representing_peak: 0,
+					peak_volume: None,
+				},
+			)]),
+		),
+	));
+
+	let tag_re_read = dump_and_re_read(&tag, WriteOptions::default().use_id3v23(true));
+
+	assert_eq!(tag_re_read.frames.len(), 0);
+}
+
+#[test]
+fn change_text_encoding_on_id3v23_save() {
+	let mut tag = Id3v2Tag::new();
+
+	// UTF-16 BE is not supported in ID3v2.3
+	tag.insert(Frame::Text(TextInformationFrame::new(
+		FrameId::Valid(Cow::from("TFOO")),
+		TextEncoding::UTF16BE,
+		String::from("Foo"),
+	)));
+
+	let tag_re_read = dump_and_re_read(&tag, WriteOptions::default().use_id3v23(true));
+
+	let frame = tag_re_read
+		.get(&FrameId::Valid(Cow::Borrowed("TFOO")))
+		.unwrap();
+	match frame {
+		Frame::Text(frame) => {
+			assert_eq!(frame.encoding, TextEncoding::UTF16);
+			assert_eq!(frame.value, "Foo");
+		},
+		_ => panic!("Expected a TextInformationFrame"),
+	}
+}
+
+#[test]
+fn split_tdor_on_id3v23_save() {
+	let mut tag = Id3v2Tag::new();
+
+	// ID3v2.3 ONLY supports the original release year.
+	// This will be written as a TORY frame. Lofty just automatically upgrades it to a TDOR
+	// when reading it back.
+	tag.insert(Frame::Timestamp(TimestampFrame::new(
+		FrameId::Valid(Cow::Borrowed("TDOR")),
+		TextEncoding::UTF8,
+		Timestamp {
+			year: 2024,
+			month: Some(6),
+			day: Some(3),
+			hour: Some(14),
+			minute: Some(8),
+			second: Some(49),
+		},
+	)));
+
+	let tag_re_read = dump_and_re_read(&tag, WriteOptions::default().use_id3v23(true));
+
+	let frame = tag_re_read
+		.get(&FrameId::Valid(Cow::Borrowed("TDOR")))
+		.unwrap();
+	match frame {
+		Frame::Timestamp(frame) => {
+			assert_eq!(frame.encoding, TextEncoding::UTF16);
+			assert_eq!(frame.timestamp.year, 2024);
+			assert_eq!(frame.timestamp.month, None);
+			assert_eq!(frame.timestamp.day, None);
+			assert_eq!(frame.timestamp.hour, None);
+			assert_eq!(frame.timestamp.minute, None);
+			assert_eq!(frame.timestamp.second, None);
+		},
+		_ => panic!("Expected a TimestampFrame"),
+	}
+}
+
+#[test]
+fn split_tdrc_on_id3v23_save() {
+	let mut tag = Id3v2Tag::new();
+
+	// TDRC gets split into 3 frames in ID3v2.3:
+	//
+	// TYER: YYYY
+	// TDAT: DDMM
+	// TIME: HHMM
+	tag.insert(Frame::Timestamp(TimestampFrame::new(
+		FrameId::Valid(Cow::Borrowed("TDRC")),
+		TextEncoding::UTF8,
+		Timestamp {
+			year: 2024,
+			month: Some(6),
+			day: Some(3),
+			hour: Some(14),
+			minute: Some(8),
+			second: None, // Seconds are not supported in ID3v2.3 TIME
+		},
+	)));
+
+	let tag_re_read = dump_and_re_read(&tag, WriteOptions::default().use_id3v23(true));
+
+	// First, check the default behavior which should return the same TDRC frame
+	let frame = tag_re_read
+		.get(&FrameId::Valid(Cow::Borrowed("TDRC")))
+		.unwrap();
+
+	match frame {
+		Frame::Timestamp(frame) => {
+			assert_eq!(frame.encoding, TextEncoding::UTF16);
+			assert_eq!(frame.timestamp.year, 2024);
+			assert_eq!(frame.timestamp.month, Some(6));
+			assert_eq!(frame.timestamp.day, Some(3));
+			assert_eq!(frame.timestamp.hour, Some(14));
+			assert_eq!(frame.timestamp.minute, Some(8));
+		},
+		_ => panic!("Expected a TimestampFrame"),
+	}
+
+	// Now, re-read with implicit_conversions off, which retains the split frames
+	let mut bytes = Cursor::new(Vec::new());
+	tag_re_read
+		.dump_to(&mut bytes, WriteOptions::default().use_id3v23(true))
+		.unwrap();
+
+	let tag_re_read = read_tag_with_options(
+		&bytes.into_inner(),
+		ParseOptions::new()
+			.parsing_mode(ParsingMode::Strict)
+			.implicit_conversions(false),
+	);
+
+	let year = tag_re_read
+		.get_text(&FrameId::Valid(Cow::Borrowed("TYER")))
+		.expect("Expected TYER frame");
+	assert_eq!(year, "2024");
+
+	let date = tag_re_read
+		.get_text(&FrameId::Valid(Cow::Borrowed("TDAT")))
+		.expect("Expected TDAT frame");
+	assert_eq!(date, "0306");
+
+	let time = tag_re_read
+		.get_text(&FrameId::Valid(Cow::Borrowed("TIME")))
+		.expect("Expected TIME frame");
+	assert_eq!(time, "1408");
 }
