@@ -5,19 +5,21 @@ mod r#ref;
 pub(crate) mod write;
 
 use super::AtomIdent;
+use crate::config::{global_options, WriteOptions};
 use crate::error::LoftyError;
 use crate::mp4::ilst::atom::AtomDataStorage;
 use crate::picture::{Picture, PictureType, TOMBSTONE_PICTURE};
-use crate::tag::item::{ItemKey, ItemValue, TagItem};
-use crate::tag::{try_parse_year, Tag, TagType};
-use crate::traits::{Accessor, MergeTag, SplitTag, TagExt};
+use crate::tag::companion_tag::CompanionTag;
+use crate::tag::{
+	try_parse_year, Accessor, ItemKey, ItemValue, MergeTag, SplitTag, Tag, TagExt, TagItem, TagType,
+};
+use crate::util::flag_item;
+use crate::util::io::{FileLike, Length, Truncate};
 use atom::{AdvisoryRating, Atom, AtomData};
 
 use std::borrow::Cow;
-use std::fs::{File, OpenOptions};
 use std::io::Write;
 use std::ops::Deref;
-use std::path::Path;
 
 use lofty_attr::tag;
 
@@ -91,7 +93,7 @@ impl Ilst {
 	///
 	/// ```rust
 	/// use lofty::mp4::Ilst;
-	/// use lofty::TagExt;
+	/// use lofty::tag::TagExt;
 	///
 	/// let ilst_tag = Ilst::new();
 	/// assert!(ilst_tag.is_empty());
@@ -106,7 +108,7 @@ impl Ilst {
 	///
 	/// ```rust
 	/// use lofty::mp4::{AtomIdent, Ilst};
-	/// use lofty::Accessor;
+	/// use lofty::tag::Accessor;
 	///
 	/// let mut ilst = Ilst::new();
 	/// ilst.set_title(String::from("Foo title"));
@@ -174,7 +176,7 @@ impl Ilst {
 	///
 	/// ```rust
 	/// use lofty::mp4::{Atom, AtomData, AtomIdent, Ilst};
-	/// use lofty::Accessor;
+	/// use lofty::tag::Accessor;
 	///
 	/// const TITLE_IDENTIFIER: AtomIdent = AtomIdent::Fourcc(*b"\xa9nam");
 	///
@@ -201,7 +203,7 @@ impl Ilst {
 	///
 	/// ```rust
 	/// use lofty::mp4::{Atom, AtomData, AtomIdent, Ilst};
-	/// use lofty::Accessor;
+	/// use lofty::tag::Accessor;
 	///
 	/// const TITLE_IDENTIFIER: AtomIdent = AtomIdent::Fourcc(*b"\xa9nam");
 	///
@@ -249,7 +251,8 @@ impl Ilst {
 	///
 	/// ```rust
 	/// use lofty::mp4::Ilst;
-	/// use lofty::{MimeType, Picture, PictureType, TagExt};
+	/// use lofty::picture::{MimeType, Picture, PictureType};
+	/// use lofty::tag::TagExt;
 	///
 	/// let mut ilst = Ilst::new();
 	///
@@ -292,7 +295,8 @@ impl Ilst {
 	///
 	/// ```rust
 	/// use lofty::mp4::Ilst;
-	/// use lofty::{MimeType, Picture, PictureType, TagExt};
+	/// use lofty::picture::{MimeType, Picture, PictureType};
+	/// use lofty::tag::TagExt;
 	///
 	/// let mut ilst = Ilst::new();
 	///
@@ -515,6 +519,11 @@ impl TagExt for Ilst {
 	type Err = LoftyError;
 	type RefKey<'a> = &'a AtomIdent<'a>;
 
+	#[inline]
+	fn tag_type(&self) -> TagType {
+		TagType::Mp4Ilst
+	}
+
 	fn len(&self) -> usize {
 		self.atoms.len()
 	}
@@ -527,25 +536,25 @@ impl TagExt for Ilst {
 		self.atoms.is_empty()
 	}
 
-	fn save_to_path<P: AsRef<Path>>(&self, path: P) -> std::result::Result<(), Self::Err> {
-		let mut f = OpenOptions::new().read(true).write(true).open(path)?;
-		self.save_to(&mut f)
+	fn save_to<F>(
+		&self,
+		file: &mut F,
+		write_options: WriteOptions,
+	) -> std::result::Result<(), Self::Err>
+	where
+		F: FileLike,
+		LoftyError: From<<F as Truncate>::Error>,
+		LoftyError: From<<F as Length>::Error>,
+	{
+		self.as_ref().write_to(file, write_options)
 	}
 
-	fn save_to(&self, file: &mut File) -> std::result::Result<(), Self::Err> {
-		self.as_ref().write_to(file)
-	}
-
-	fn dump_to<W: Write>(&self, writer: &mut W) -> std::result::Result<(), Self::Err> {
-		self.as_ref().dump_to(writer)
-	}
-
-	fn remove_from_path<P: AsRef<Path>>(&self, path: P) -> std::result::Result<(), Self::Err> {
-		TagType::Mp4Ilst.remove_from_path(path)
-	}
-
-	fn remove_from(&self, file: &mut File) -> std::result::Result<(), Self::Err> {
-		TagType::Mp4Ilst.remove_from(file)
+	fn dump_to<W: Write>(
+		&self,
+		writer: &mut W,
+		write_options: WriteOptions,
+	) -> std::result::Result<(), Self::Err> {
+		self.as_ref().dump_to(writer, write_options)
 	}
 
 	fn clear(&mut self) {
@@ -592,7 +601,7 @@ impl SplitTag for Ilst {
 					ItemValue::Text(text)
 				},
 				// We have to special case track/disc numbers since they are stored together
-				AtomData::Unknown { code: 0, data } if data.len() >= 6 => {
+				AtomData::Unknown { code: 0, data } if Vec::len(data) >= 6 => {
 					if let AtomIdent::Fourcc(ref fourcc) = ident {
 						match fourcc {
 							b"trkn" => {
@@ -646,6 +655,11 @@ impl SplitTag for Ilst {
 			false // Atom consumed
 		});
 
+		if let Some(rating) = self.advisory_rating() {
+			tag.insert_text(ItemKey::ParentalAdvisory, rating.as_u8().to_string());
+			let _ = self.remove(&ADVISORY_RATING);
+		}
+
 		(SplitTagRemainder(self), tag)
 	}
 }
@@ -698,23 +712,38 @@ impl MergeTag for SplitTagRemainder {
 					ItemKey::DiscNumber => convert_to_uint(&mut discs.0, text.as_str()),
 					ItemKey::DiscTotal => convert_to_uint(&mut discs.1, text.as_str()),
 					ItemKey::FlagCompilation | ItemKey::FlagPodcast => {
-						let Ok(num) = text.as_str().parse::<u8>() else {
+						let Some(data) = flag_item(text.as_str()) else {
 							continue;
 						};
 
-						let data = match num {
-							0 => false,
-							1 => true,
-							_ => {
-								// Ignore all other, unexpected values
-								continue;
-							},
-						};
 						merged.atoms.push(Atom {
 							ident: ident.into_owned(),
 							data: AtomDataStorage::Single(AtomData::Bool(data)),
 						})
 					},
+					ItemKey::ParentalAdvisory => {
+						let Ok(rating) = text.parse::<u8>() else {
+							log::warn!(
+								"Parental advisory rating is not a number: {}, discarding",
+								text
+							);
+							continue;
+						};
+
+						let Ok(parsed_rating) = AdvisoryRating::try_from(rating) else {
+							log::warn!(
+								"Parental advisory rating is out of range: {rating}, discarding"
+							);
+							continue;
+						};
+
+						merged.atoms.push(Atom {
+							ident: ident.into_owned(),
+							data: AtomDataStorage::Single(AtomData::SignedInteger(i32::from(
+								parsed_rating.as_u8(),
+							))),
+						})
+					},
 					_ => merged.atoms.push(Atom {
 						ident: ident.into_owned(),
 						data: AtomDataStorage::Single(AtomData::UTF8(text)),
@@ -743,38 +772,51 @@ impl MergeTag for SplitTagRemainder {
 
 impl From<Ilst> for Tag {
 	fn from(input: Ilst) -> Self {
-		input.split_tag().1
+		let (remainder, mut tag) = input.split_tag();
+
+		if unsafe { global_options().preserve_format_specific_items } && remainder.0.len() > 0 {
+			tag.companion_tag = Some(CompanionTag::Ilst(remainder.0));
+		}
+
+		tag
 	}
 }
 
 impl From<Tag> for Ilst {
-	fn from(input: Tag) -> Self {
+	fn from(mut input: Tag) -> Self {
+		if unsafe { global_options().preserve_format_specific_items } {
+			if let Some(companion) = input.companion_tag.take().and_then(CompanionTag::ilst) {
+				return SplitTagRemainder(companion).merge_tag(input);
+			}
+		}
+
 		SplitTagRemainder::default().merge_tag(input)
 	}
 }
 
 #[cfg(test)]
 mod tests {
+	use crate::config::{ParseOptions, ParsingMode, WriteOptions};
 	use crate::mp4::ilst::atom::AtomDataStorage;
 	use crate::mp4::ilst::TITLE;
 	use crate::mp4::read::AtomReader;
 	use crate::mp4::{AdvisoryRating, Atom, AtomData, AtomIdent, Ilst, Mp4File};
+	use crate::prelude::*;
 	use crate::tag::utils::test_utils;
 	use crate::tag::utils::test_utils::read_path;
-	use crate::{
-		Accessor as _, AudioFile, ItemKey, ItemValue, ParseOptions, ParsingMode, SplitTag as _,
-		Tag, TagExt as _, TagItem, TagType,
-	};
+	use crate::tag::{ItemValue, Tag, TagItem, TagType};
+
+	use crate::picture::{MimeType, Picture, PictureType};
 	use std::io::{Cursor, Read as _, Seek as _, Write as _};
 
 	fn read_ilst(path: &str, parse_mode: ParsingMode) -> Ilst {
 		let tag = std::fs::read(path).unwrap();
-		let len = tag.len();
+		read_ilst_raw(&tag, parse_mode)
+	}
 
-		let cursor = Cursor::new(tag);
-		let mut reader = AtomReader::new(cursor, parse_mode).unwrap();
-
-		super::read::parse_ilst(&mut reader, parse_mode, len as u64).unwrap()
+	fn read_ilst_raw(bytes: &[u8], parse_mode: ParsingMode) -> Ilst {
+		let options = ParseOptions::new().parsing_mode(parse_mode);
+		read_ilst_with_options(bytes, options)
 	}
 
 	fn read_ilst_strict(path: &str) -> Ilst {
@@ -785,6 +827,15 @@ mod tests {
 		read_ilst(path, ParsingMode::BestAttempt)
 	}
 
+	fn read_ilst_with_options(bytes: &[u8], parse_options: ParseOptions) -> Ilst {
+		let len = bytes.len();
+
+		let cursor = Cursor::new(bytes);
+		let mut reader = AtomReader::new(cursor, parse_options.parsing_mode).unwrap();
+
+		super::read::parse_ilst(&mut reader, parse_options, len as u64).unwrap()
+	}
+
 	fn verify_atom(ilst: &Ilst, ident: [u8; 4], data: &AtomData) {
 		let atom = ilst.get(&AtomIdent::Fourcc(ident)).unwrap();
 		assert_eq!(atom.data().next().unwrap(), data);
@@ -848,10 +899,14 @@ mod tests {
 		let len = tag.len();
 
 		let cursor = Cursor::new(tag);
-		let mut reader = AtomReader::new(cursor, crate::ParsingMode::Strict).unwrap();
+		let mut reader = AtomReader::new(cursor, ParsingMode::Strict).unwrap();
 
-		let parsed_tag =
-			super::read::parse_ilst(&mut reader, crate::ParsingMode::Strict, len as u64).unwrap();
+		let parsed_tag = super::read::parse_ilst(
+			&mut reader,
+			ParseOptions::new().parsing_mode(ParsingMode::Strict),
+			len as u64,
+		)
+		.unwrap();
 
 		assert_eq!(expected_tag, parsed_tag);
 	}
@@ -861,15 +916,17 @@ mod tests {
 		let parsed_tag = read_ilst_strict("tests/tags/assets/ilst/test.ilst");
 
 		let mut writer = Vec::new();
-		parsed_tag.dump_to(&mut writer).unwrap();
+		parsed_tag
+			.dump_to(&mut writer, WriteOptions::default())
+			.unwrap();
 
 		let cursor = Cursor::new(&writer[8..]);
-		let mut reader = AtomReader::new(cursor, crate::ParsingMode::Strict).unwrap();
+		let mut reader = AtomReader::new(cursor, ParsingMode::Strict).unwrap();
 
 		// Remove the ilst identifier and size
 		let temp_parsed_tag = super::read::parse_ilst(
 			&mut reader,
-			crate::ParsingMode::Strict,
+			ParseOptions::new().parsing_mode(ParsingMode::Strict),
 			(writer.len() - 8) as u64,
 		)
 		.unwrap();
@@ -883,10 +940,14 @@ mod tests {
 		let len = tag.len();
 
 		let cursor = Cursor::new(tag);
-		let mut reader = AtomReader::new(cursor, crate::ParsingMode::Strict).unwrap();
+		let mut reader = AtomReader::new(cursor, ParsingMode::Strict).unwrap();
 
-		let ilst =
-			super::read::parse_ilst(&mut reader, crate::ParsingMode::Strict, len as u64).unwrap();
+		let ilst = super::read::parse_ilst(
+			&mut reader,
+			ParseOptions::new().parsing_mode(ParsingMode::Strict),
+			len as u64,
+		)
+		.unwrap();
 
 		let tag: Tag = ilst.into();
 
@@ -1003,11 +1064,11 @@ mod tests {
 			assert_eq!(old_free_size, PADDING_SIZE as u32);
 
 			let cursor = Cursor::new(ilst_bytes);
-			let mut reader = AtomReader::new(cursor, crate::ParsingMode::Strict).unwrap();
+			let mut reader = AtomReader::new(cursor, ParsingMode::Strict).unwrap();
 
 			ilst = super::read::parse_ilst(
 				&mut reader,
-				crate::ParsingMode::Strict,
+				ParseOptions::new().parsing_mode(ParsingMode::Strict),
 				ilst_bytes.len() as u64,
 			)
 			.unwrap();
@@ -1018,7 +1079,7 @@ mod tests {
 		file.rewind().unwrap();
 
 		ilst.set_title(String::from("Exactly 21 Characters"));
-		ilst.save_to(&mut file).unwrap();
+		ilst.save_to(&mut file, WriteOptions::default()).unwrap();
 
 		// Now verify the free atom
 		file.rewind().unwrap();
@@ -1073,7 +1134,7 @@ mod tests {
 			data: AtomDataStorage::Single(AtomData::UTF8(String::from("Foo artist"))),
 		});
 
-		tag.save_to(&mut file).unwrap();
+		tag.save_to(&mut file, WriteOptions::default()).unwrap();
 		file.rewind().unwrap();
 
 		let mp4_file = Mp4File::read_from(&mut file, ParseOptions::new()).unwrap();
@@ -1250,4 +1311,75 @@ mod tests {
 			&AtomData::Bool(false)
 		);
 	}
+
+	#[test]
+	fn special_items_roundtrip() {
+		let mut tag = Ilst::new();
+
+		let atom = Atom::new(
+			AtomIdent::Fourcc(*b"SMTH"),
+			AtomData::Unknown {
+				code: 0,
+				data: b"Meaningless Data".to_vec(),
+			},
+		);
+
+		tag.insert(atom.clone());
+		tag.set_artist(String::from("Foo Artist")); // Some value that we *can* represent generically
+
+		let tag: Tag = tag.into();
+
+		assert_eq!(tag.len(), 1);
+		assert_eq!(tag.artist().as_deref(), Some("Foo Artist"));
+
+		let tag: Ilst = tag.into();
+
+		assert_eq!(tag.atoms.len(), 2);
+		assert_eq!(tag.artist().as_deref(), Some("Foo Artist"));
+		assert_eq!(tag.get(&AtomIdent::Fourcc(*b"SMTH")), Some(&atom));
+
+		let mut tag_bytes = Vec::new();
+		tag.dump_to(&mut tag_bytes, WriteOptions::default())
+			.unwrap();
+
+		tag_bytes.drain(..8); // Remove the ilst identifier and size for `read_ilst`
+
+		let tag_re_read = read_ilst_raw(&tag_bytes[..], ParsingMode::Strict);
+		assert_eq!(tag, tag_re_read);
+
+		// Now write from `Tag`
+		let tag: Tag = tag.into();
+
+		let mut tag_bytes = Vec::new();
+		tag.dump_to(&mut tag_bytes, WriteOptions::default())
+			.unwrap();
+
+		tag_bytes.drain(..8); // Remove the ilst identifier and size for `read_ilst`
+
+		let generic_tag_re_read = read_ilst_raw(&tag_bytes[..], ParsingMode::Strict);
+		assert_eq!(tag_re_read, generic_tag_re_read);
+	}
+
+	#[test]
+	fn skip_reading_cover_art() {
+		let p = Picture::new_unchecked(
+			PictureType::CoverFront,
+			Some(MimeType::Jpeg),
+			None,
+			std::iter::repeat(0).take(50).collect::<Vec<u8>>(),
+		);
+
+		let mut tag = Tag::new(TagType::Mp4Ilst);
+		tag.push_picture(p);
+
+		tag.set_artist(String::from("Foo artist"));
+
+		let mut writer = Vec::new();
+		tag.dump_to(&mut writer, WriteOptions::new()).unwrap();
+
+		// Skip `ilst` header
+		let ilst = read_ilst_with_options(&writer[8..], ParseOptions::new().read_cover_art(false));
+		assert_eq!(ilst.len(), 1); // Artist, no picture
+		assert!(ilst.artist().is_some());
+	}
 }
