@@ -2,10 +2,11 @@
 // Reference Conversions
 // *********************
 
-use crate::error::Result;
+use crate::config::WriteOptions;
+use crate::error::{LoftyError, Result};
 use crate::mp4::{Atom, AtomData, AtomIdent, Ilst};
+use crate::util::io::{FileLike, Length, Truncate};
 
-use std::fs::File;
 use std::io::Write;
 
 impl Ilst {
@@ -24,11 +25,20 @@ impl<'a, I: 'a> IlstRef<'a, I>
 where
 	I: IntoIterator<Item = &'a AtomData>,
 {
-	pub(crate) fn write_to(&mut self, file: &mut File) -> Result<()> {
-		super::write::write_to(file, self)
+	pub(crate) fn write_to<F>(&mut self, file: &mut F, write_options: WriteOptions) -> Result<()>
+	where
+		F: FileLike,
+		LoftyError: From<<F as Truncate>::Error>,
+		LoftyError: From<<F as Length>::Error>,
+	{
+		super::write::write_to(file, self, write_options)
 	}
 
-	pub(crate) fn dump_to<W: Write>(&mut self, writer: &mut W) -> Result<()> {
+	pub(crate) fn dump_to<W: Write>(
+		&mut self,
+		writer: &mut W,
+		_write_options: WriteOptions,
+	) -> Result<()> {
 		let temp = super::write::build_ilst(&mut self.atoms)?;
 		writer.write_all(&temp)?;
 
