@@ -1,5 +1,6 @@
 use super::r#ref::IlstRef;
-use crate::error::{FileEncodingError, Result};
+use crate::config::{ParseOptions, WriteOptions};
+use crate::error::{FileEncodingError, LoftyError, Result};
 use crate::file::FileType;
 use crate::macros::{decode_err, err, try_vec};
 use crate::mp4::atom_info::{AtomIdent, AtomInfo, ATOM_HEADER_LEN, FOURCC_LEN};
@@ -8,9 +9,9 @@ use crate::mp4::read::{atom_tree, meta_is_full, nested_atom, verify_mp4, AtomRea
 use crate::mp4::write::{AtomWriter, AtomWriterCompanion, ContextualAtom};
 use crate::mp4::AtomData;
 use crate::picture::{MimeType, Picture};
-use crate::probe::ParseOptions;
+use crate::util::alloc::VecFallibleCapacity;
+use crate::util::io::{FileLike, Length, Truncate};
 
-use std::fs::File;
 use std::io::{Cursor, Seek, SeekFrom, Write};
 
 use byteorder::{BigEndian, ReadBytesExt, WriteBytesExt};
@@ -18,24 +19,30 @@ use byteorder::{BigEndian, ReadBytesExt, WriteBytesExt};
 // A "full" atom is a traditional length + identifier, followed by a version (1) and flags (3)
 const FULL_ATOM_SIZE: u64 = ATOM_HEADER_LEN + 4;
 const HDLR_SIZE: u64 = ATOM_HEADER_LEN + 25;
-const DEFAULT_PADDING_SIZE: u32 = 1024;
 
 // TODO: We are forcing the use of ParseOptions::DEFAULT_PARSING_MODE. This is not good. It should be caller-specified.
-pub(crate) fn write_to<'a, I: 'a>(data: &mut File, tag: &mut IlstRef<'a, I>) -> Result<()>
+pub(crate) fn write_to<'a, F, I>(
+	file: &mut F,
+	tag: &mut IlstRef<'a, I>,
+	write_options: WriteOptions,
+) -> Result<()>
 where
-	I: IntoIterator<Item = &'a AtomData>,
+	F: FileLike,
+	LoftyError: From<<F as Truncate>::Error>,
+	LoftyError: From<<F as Length>::Error>,
+	I: IntoIterator<Item = &'a AtomData> + 'a,
 {
 	log::debug!("Attempting to write `ilst` tag to file");
 
 	// Create a temporary `AtomReader`, just to verify that this is a valid MP4 file
-	let mut reader = AtomReader::new(data, ParseOptions::DEFAULT_PARSING_MODE)?;
+	let mut reader = AtomReader::new(file, ParseOptions::DEFAULT_PARSING_MODE)?;
 	verify_mp4(&mut reader)?;
 
 	// Now we can just read the entire file into memory
-	let data = reader.into_inner();
-	data.rewind()?;
+	let file = reader.into_inner();
+	file.rewind()?;
 
-	let mut atom_writer = AtomWriter::new_from_file(data, ParseOptions::DEFAULT_PARSING_MODE)?;
+	let mut atom_writer = AtomWriter::new_from_file(file, ParseOptions::DEFAULT_PARSING_MODE)?;
 
 	let Some(moov) = atom_writer.find_contextual_atom(*b"moov") else {
 		return Err(FileEncodingError::new(
@@ -126,6 +133,7 @@ where
 					&mut new_udta_size,
 					ilst,
 					remove_tag,
+					write_options,
 				)?
 			},
 			// We have to create the `meta` atom
@@ -194,7 +202,7 @@ where
 
 	drop(write_handle);
 
-	atom_writer.save_to(data)?;
+	atom_writer.save_to(file)?;
 
 	Ok(())
 }
@@ -207,6 +215,7 @@ fn save_to_existing(
 	new_udta_size: &mut u64,
 	ilst: Vec<u8>,
 	remove_tag: bool,
+	write_options: WriteOptions,
 ) -> Result<()> {
 	let mut replacement;
 	let range;
@@ -248,21 +257,35 @@ fn save_to_existing(
 
 			// Check for one directly before the `ilst` atom
 			if ilst_idx > 0 {
-				let previous_atom = &tree[ilst_idx - 1];
+				let mut i = ilst_idx;
+				while i != 0 {
+					let atom = &tree[i - 1];
+					if atom.ident != AtomIdent::Fourcc(*b"free") {
+						break;
+					}
 
-				if previous_atom.ident == AtomIdent::Fourcc(*b"free") {
-					range_start = previous_atom.start;
-					available_space += previous_atom.len;
+					available_space += atom.len;
+					range_start = atom.start;
+					i -= 1;
 				}
+
+				log::trace!("Found {} preceding `free` atoms", ilst_idx - i)
 			}
 
 			// And after
 			if ilst_idx != tree.len() - 1 {
-				let next_atom = &tree[ilst_idx + 1];
+				let mut i = ilst_idx;
+				while i < tree.len() - 1 {
+					let atom = &tree[i + 1];
+					if atom.ident != AtomIdent::Fourcc(*b"free") {
+						break;
+					}
 
-				if next_atom.ident == AtomIdent::Fourcc(*b"free") {
-					available_space += next_atom.len;
+					available_space += atom.len;
+					i += 1;
 				}
+
+				log::trace!("Found {} succeeding `free` atoms", i - ilst_idx)
 			}
 
 			let ilst_len = ilst.len() as u64;
@@ -303,7 +326,8 @@ fn save_to_existing(
 		log::trace!("Tag size changed, attempting to avoid offset update");
 
 		let mut ilst_writer = Cursor::new(replacement);
-		let (atom_size_difference, padding_size) = pad_atom(&mut ilst_writer, difference)?;
+		let (atom_size_difference, padding_size) =
+			pad_atom(&mut ilst_writer, difference, write_options)?;
 
 		replacement = ilst_writer.into_inner();
 		new_meta_size += padding_size;
@@ -339,7 +363,11 @@ fn save_to_existing(
 	Ok(())
 }
 
-fn pad_atom<W>(writer: &mut W, mut atom_size_difference: i64) -> Result<(i64, u64)>
+fn pad_atom<W>(
+	writer: &mut W,
+	mut atom_size_difference: i64,
+	write_options: WriteOptions,
+) -> Result<(i64, u64)>
 where
 	W: Write + Seek,
 {
@@ -354,7 +382,7 @@ where
 
 	let padding_size: u64;
 	let diff_abs = atom_size_difference.abs();
-	if diff_abs >= 8 {
+	if diff_abs >= ATOM_HEADER_LEN as i64 {
 		log::trace!(
 			"Avoiding offset update, padding atom with {} bytes",
 			diff_abs
@@ -365,19 +393,26 @@ where
 		write_free_atom(writer, diff_abs as u32)?;
 		atom_size_difference = 0;
 		padding_size = diff_abs as u64;
-	} else {
-		log::trace!(
-			"Cannot avoid offset update, padding atom with {} bytes",
-			DEFAULT_PADDING_SIZE
-		);
 
-		// Otherwise, we'll have to just pad the default amount,
-		// and update the offsets.
-		write_free_atom(writer, DEFAULT_PADDING_SIZE)?;
-		atom_size_difference += i64::from(DEFAULT_PADDING_SIZE);
-		padding_size = u64::from(DEFAULT_PADDING_SIZE);
+		return Ok((atom_size_difference, padding_size));
 	}
 
+	let Some(preferred_padding) = write_options.preferred_padding else {
+		log::trace!("Cannot avoid offset update, not padding atom");
+		return Ok((atom_size_difference, 0));
+	};
+
+	log::trace!(
+		"Cannot avoid offset update, padding atom with {} bytes",
+		preferred_padding
+	);
+
+	// Otherwise, we'll have to just pad the default amount,
+	// and update the offsets.
+	write_free_atom(writer, preferred_padding)?;
+	atom_size_difference += i64::from(preferred_padding);
+	padding_size = u64::from(preferred_padding);
+
 	Ok((atom_size_difference, padding_size))
 }
 
@@ -506,7 +541,7 @@ fn create_udta(ilst: &[u8]) -> Result<Vec<u8>> {
 
 	// `udta` + `meta` + `hdlr` + `ilst`
 	let capacity = ATOM_HEADER_LEN + FULL_ATOM_SIZE + HDLR_SIZE + ilst.len() as u64;
-	let mut buf = Vec::with_capacity(capacity as usize);
+	let mut buf = Vec::try_with_capacity_stable(capacity as usize)?;
 
 	buf.write_all(&UDTA_HEADER)?;
 
@@ -557,11 +592,9 @@ fn create_meta(writer: &AtomWriter, ilst: &[u8]) -> Result<()> {
 	Ok(())
 }
 
-pub(super) fn build_ilst<'a, I: 'a>(
-	atoms: &mut dyn Iterator<Item = AtomRef<'a, I>>,
-) -> Result<Vec<u8>>
+pub(super) fn build_ilst<'a, I>(atoms: &mut dyn Iterator<Item = AtomRef<'a, I>>) -> Result<Vec<u8>>
 where
-	I: IntoIterator<Item = &'a AtomData>,
+	I: IntoIterator<Item = &'a AtomData> + 'a,
 {
 	log::debug!("Building `ilst` atom");
 
@@ -636,9 +669,9 @@ where
 	Ok(())
 }
 
-fn write_atom_data<'a, I: 'a>(data: I, writer: &mut AtomWriterCompanion<'_>) -> Result<()>
+fn write_atom_data<'a, I>(data: I, writer: &mut AtomWriterCompanion<'_>) -> Result<()>
 where
-	I: IntoIterator<Item = &'a AtomData>,
+	I: IntoIterator<Item = &'a AtomData> + 'a,
 {
 	for value in data {
 		match value {
