@@ -1,3 +1,4 @@
+use crate::config::WriteOptions;
 use crate::error::{LoftyError, Result};
 use crate::file::FileType;
 use crate::macros::err;
@@ -5,15 +6,15 @@ use crate::ogg::picture_storage::OggPictureStorage;
 use crate::ogg::write::OGGFormat;
 use crate::picture::{Picture, PictureInformation};
 use crate::probe::Probe;
-use crate::tag::item::{ItemKey, ItemValue, TagItem};
-use crate::tag::{try_parse_year, Tag, TagType};
-use crate::traits::{Accessor, MergeTag, SplitTag, TagExt};
+use crate::tag::{
+	try_parse_year, Accessor, ItemKey, ItemValue, MergeTag, SplitTag, Tag, TagExt, TagItem, TagType,
+};
+use crate::util::flag_item;
+use crate::util::io::{FileLike, Length, Truncate};
 
 use std::borrow::Cow;
-use std::fs::File;
 use std::io::Write;
 use std::ops::Deref;
-use std::path::Path;
 
 use lofty_attr::tag;
 
@@ -78,7 +79,7 @@ impl VorbisComments {
 	///
 	/// ```rust
 	/// use lofty::ogg::VorbisComments;
-	/// use lofty::TagExt;
+	/// use lofty::tag::TagExt;
 	///
 	/// let vorbis_comments_tag = VorbisComments::new();
 	/// assert!(vorbis_comments_tag.is_empty());
@@ -143,7 +144,7 @@ impl VorbisComments {
 	///
 	/// ```rust
 	/// use lofty::ogg::VorbisComments;
-	/// use lofty::TagExt;
+	/// use lofty::tag::TagExt;
 	///
 	/// let mut vorbis_comments = VorbisComments::default();
 	///
@@ -432,6 +433,11 @@ impl TagExt for VorbisComments {
 	type Err = LoftyError;
 	type RefKey<'a> = &'a str;
 
+	#[inline]
+	fn tag_type(&self) -> TagType {
+		TagType::VorbisComments
+	}
+
 	fn len(&self) -> usize {
 		self.items.len() + self.pictures.len()
 	}
@@ -454,13 +460,22 @@ impl TagExt for VorbisComments {
 	/// * The file does not contain valid packets
 	/// * [`PictureInformation::from_picture`]
 	/// * [`std::io::Error`]
-	fn save_to(&self, file: &mut File) -> std::result::Result<(), Self::Err> {
+	fn save_to<F>(
+		&self,
+		file: &mut F,
+		write_options: WriteOptions,
+	) -> std::result::Result<(), Self::Err>
+	where
+		F: FileLike,
+		LoftyError: From<<F as Truncate>::Error>,
+		LoftyError: From<<F as Length>::Error>,
+	{
 		VorbisCommentsRef {
-			vendor: self.vendor.as_str(),
+			vendor: Cow::from(self.vendor.as_str()),
 			items: self.items.iter().map(|(k, v)| (k.as_str(), v.as_str())),
 			pictures: self.pictures.iter().map(|(p, i)| (p, *i)),
 		}
-		.write_to(file)
+		.write_to(file, write_options)
 	}
 
 	/// Dumps the tag to a writer
@@ -472,21 +487,17 @@ impl TagExt for VorbisComments {
 	///
 	/// * [`PictureInformation::from_picture`]
 	/// * [`std::io::Error`]
-	fn dump_to<W: Write>(&self, writer: &mut W) -> std::result::Result<(), Self::Err> {
+	fn dump_to<W: Write>(
+		&self,
+		writer: &mut W,
+		write_options: WriteOptions,
+	) -> std::result::Result<(), Self::Err> {
 		VorbisCommentsRef {
-			vendor: self.vendor.as_str(),
+			vendor: Cow::from(self.vendor.as_str()),
 			items: self.items.iter().map(|(k, v)| (k.as_str(), v.as_str())),
 			pictures: self.pictures.iter().map(|(p, i)| (p, *i)),
 		}
-		.dump_to(writer)
-	}
-
-	fn remove_from_path<P: AsRef<Path>>(&self, path: P) -> std::result::Result<(), Self::Err> {
-		TagType::VorbisComments.remove_from_path(path)
-	}
-
-	fn remove_from(&self, file: &mut File) -> std::result::Result<(), Self::Err> {
-		TagType::VorbisComments.remove_from(file)
+		.dump_to(writer, write_options)
 	}
 
 	fn clear(&mut self) {
@@ -565,10 +576,19 @@ impl MergeTag for SplitTagRemainder {
 			let item_value = item.item_value;
 
 			// Discard binary items, as they are not allowed in Vorbis comments
-			let (ItemValue::Text(val) | ItemValue::Locator(val)) = item_value else {
+			let (ItemValue::Text(mut val) | ItemValue::Locator(mut val)) = item_value else {
 				continue;
 			};
 
+			// Normalize flag items
+			if item_key == ItemKey::FlagCompilation {
+				let Some(flag) = flag_item(&val) else {
+					continue;
+				};
+
+				val = u8::from(flag).to_string();
+			}
+
 			let key;
 			match item_key {
 				ItemKey::Unknown(unknown) => {
@@ -614,7 +634,7 @@ where
 	II: Iterator<Item = (&'a str, &'a str)>,
 	IP: Iterator<Item = (&'a Picture, PictureInformation)>,
 {
-	pub vendor: &'a str,
+	pub vendor: Cow<'a, str>,
 	pub items: II,
 	pub pictures: IP,
 }
@@ -625,7 +645,12 @@ where
 	IP: Iterator<Item = (&'a Picture, PictureInformation)>,
 {
 	#[allow(clippy::shadow_unrelated)]
-	pub(crate) fn write_to(&mut self, file: &mut File) -> Result<()> {
+	pub(crate) fn write_to<F>(&mut self, file: &mut F, write_options: WriteOptions) -> Result<()>
+	where
+		F: FileLike,
+		LoftyError: From<<F as Truncate>::Error>,
+		LoftyError: From<<F as Length>::Error>,
+	{
 		let probe = Probe::new(file).guess_file_type()?;
 		let f_ty = probe.file_type();
 
@@ -638,17 +663,20 @@ where
 
 		// FLAC has its own special writing needs :)
 		if file_type == FileType::Flac {
-			return crate::flac::write::write_to_inner(file, self);
+			return crate::flac::write::write_to_inner(file, self, write_options);
 		}
 
 		let (format, header_packet_count) = OGGFormat::from_filetype(file_type);
 
-		super::write::write(file, self, format, header_packet_count)
+		super::write::write(file, self, format, header_packet_count, write_options)
 	}
 
-	pub(crate) fn dump_to<W: Write>(&mut self, writer: &mut W) -> Result<()> {
-		let metadata_packet =
-			super::write::create_metadata_packet(self, &[], self.vendor.as_bytes(), false)?;
+	pub(crate) fn dump_to<W: Write>(
+		&mut self,
+		writer: &mut W,
+		_write_options: WriteOptions,
+	) -> Result<()> {
+		let metadata_packet = super::write::create_metadata_packet(self, &[], false)?;
 		writer.write_all(&metadata_packet)?;
 		Ok(())
 	}
@@ -680,16 +708,22 @@ pub(crate) fn create_vorbis_comments_ref(
 
 #[cfg(test)]
 mod tests {
+	use crate::config::{ParseOptions, ParsingMode, WriteOptions};
 	use crate::ogg::{OggPictureStorage, VorbisComments};
-	use crate::{
-		ItemKey, ItemValue, MergeTag as _, ParsingMode, SplitTag as _, Tag, TagExt as _, TagItem,
-		TagType,
-	};
+	use crate::picture::{MimeType, Picture, PictureType};
+	use crate::prelude::*;
+	use crate::tag::{ItemValue, Tag, TagItem, TagType};
+	use std::io::Cursor;
 
 	fn read_tag(tag: &[u8]) -> VorbisComments {
 		let mut reader = std::io::Cursor::new(tag);
 
-		crate::ogg::read::read_comments(&mut reader, tag.len() as u64, ParsingMode::Strict).unwrap()
+		crate::ogg::read::read_comments(
+			&mut reader,
+			tag.len() as u64,
+			ParseOptions::new().parsing_mode(ParsingMode::Strict),
+		)
+		.unwrap()
 	}
 
 	#[test]
@@ -721,7 +755,9 @@ mod tests {
 		parsed_tag.vendor = String::new();
 
 		let mut writer = Vec::new();
-		parsed_tag.dump_to(&mut writer).unwrap();
+		parsed_tag
+			.dump_to(&mut writer, WriteOptions::default())
+			.unwrap();
 
 		let temp_parsed_tag = read_tag(&writer);
 
@@ -864,4 +900,35 @@ mod tests {
 			assert_eq!(Some("Cmin"), vorbis_comments.get("INITIALKEY"));
 		}
 	}
+
+	#[test]
+	fn skip_reading_cover_art() {
+		let p = Picture::new_unchecked(
+			PictureType::CoverFront,
+			Some(MimeType::Jpeg),
+			None,
+			std::iter::repeat(0).take(50).collect::<Vec<u8>>(),
+		);
+
+		let mut tag = Tag::new(TagType::VorbisComments);
+		tag.push_picture(p);
+
+		tag.set_artist(String::from("Foo artist"));
+
+		let mut writer = Vec::new();
+		tag.dump_to(&mut writer, WriteOptions::new()).unwrap();
+
+		let mut reader = Cursor::new(&writer);
+		let tag = crate::ogg::read::read_comments(
+			&mut reader,
+			writer.len() as u64,
+			ParseOptions::new()
+				.parsing_mode(ParsingMode::Strict)
+				.read_cover_art(false),
+		)
+		.unwrap();
+
+		assert_eq!(tag.pictures().len(), 0); // Artist, no picture
+		assert!(tag.artist().is_some());
+	}
 }
