@@ -1,19 +1,32 @@
+//! Utilities for generic tag handling
+
+mod accessor;
+pub(crate) mod companion_tag;
 pub(crate) mod item;
+pub mod items;
+mod split_merge_tag;
+mod tag_ext;
+mod tag_type;
 pub(crate) mod utils;
 
+use crate::config::WriteOptions;
 use crate::error::{LoftyError, Result};
-use crate::file::FileType;
 use crate::macros::err;
 use crate::picture::{Picture, PictureType};
 use crate::probe::Probe;
-use crate::traits::{Accessor, MergeTag, SplitTag, TagExt};
-use item::{ItemKey, ItemValue, TagItem};
+use crate::util::io::{FileLike, Length, Truncate};
 
 use std::borrow::Cow;
-use std::fs::{File, OpenOptions};
 use std::io::Write;
 use std::path::Path;
 
+// Exports
+pub use accessor::Accessor;
+pub use item::{ItemKey, ItemValue, TagItem};
+pub use split_merge_tag::{MergeTag, SplitTag};
+pub use tag_ext::TagExt;
+pub use tag_type::TagType;
+
 macro_rules! impl_accessor {
 	($($item_key:ident => $name:tt),+) => {
 		paste::paste! {
@@ -56,7 +69,7 @@ macro_rules! impl_accessor {
 /// Accessing common items
 ///
 /// ```rust
-/// use lofty::{Accessor, Tag, TagType};
+/// use lofty::tag::{Accessor, Tag, TagType};
 ///
 /// let tag = Tag::new(TagType::Id3v2);
 ///
@@ -71,7 +84,7 @@ macro_rules! impl_accessor {
 /// Getting an item of a known type
 ///
 /// ```rust
-/// use lofty::{ItemKey, Tag, TagType};
+/// use lofty::tag::{ItemKey, Tag, TagType};
 ///
 /// let tag = Tag::new(TagType::Id3v2);
 ///
@@ -86,7 +99,7 @@ macro_rules! impl_accessor {
 ///
 /// ```rust
 /// use lofty::id3::v2::Id3v2Tag;
-/// use lofty::{Tag, TagType};
+/// use lofty::tag::{Tag, TagType};
 ///
 /// // Converting between formats is as simple as an `into` call.
 /// // However, such conversions can potentially be *very* lossy.
@@ -99,6 +112,7 @@ pub struct Tag {
 	tag_type: TagType,
 	pub(crate) pictures: Vec<Picture>,
 	pub(crate) items: Vec<TagItem>,
+	pub(crate) companion_tag: Option<companion_tag::CompanionTag>,
 }
 
 #[must_use]
@@ -215,15 +229,64 @@ impl Tag {
 			tag_type,
 			pictures: Vec::new(),
 			items: Vec::new(),
+			companion_tag: None,
 		}
 	}
 
 	/// Change the [`TagType`], remapping all items
+	///
+	/// NOTE: If any format-specific items are present, they will be removed.
+	///       See [`GlobalOptions::preserve_format_specific_items`].
+	///
+	/// # Examples
+	///
+	/// ```rust
+	/// use lofty::tag::{Accessor, Tag, TagExt, TagType};
+	///
+	/// let mut tag = Tag::new(TagType::Id3v2);
+	/// tag.set_album(String::from("Album"));
+	///
+	/// // ID3v2 supports the album tag
+	/// assert_eq!(tag.len(), 1);
+	///
+	/// // But AIFF text chunks do not, the item will be lost
+	/// tag.re_map(TagType::AiffText);
+	/// assert!(tag.is_empty());
+	/// ```
+	///
+	/// [`GlobalOptions::preserve_format_specific_items`]: crate::config::GlobalOptions::preserve_format_specific_items
 	pub fn re_map(&mut self, tag_type: TagType) {
+		if let Some(companion_tag) = self.companion_tag.take() {
+			log::warn!("Discarding format-specific items due to remap");
+			drop(companion_tag);
+		}
+
 		self.retain(|i| i.re_map(tag_type));
 		self.tag_type = tag_type
 	}
 
+	/// Check if the tag contains any format-specific items
+	///
+	/// See [`GlobalOptions::preserve_format_specific_items`].
+	///
+	/// # Examples
+	///
+	/// ```rust
+	/// use lofty::tag::{Accessor, Tag, TagExt, TagType};
+	///
+	/// let mut tag = Tag::new(TagType::Id3v2);
+	/// tag.set_album(String::from("Album"));
+	///
+	/// // We cannot create a tag with format-specific items.
+	/// // This must come from a conversion, such as `Id3v2Tag` -> `Tag`
+	/// assert!(!tag.has_format_specific_items());
+	/// ```
+	///
+	/// [`GlobalOptions::preserve_format_specific_items`]: crate::config::GlobalOptions::preserve_format_specific_items
+	pub fn has_format_specific_items(&self) -> bool {
+		self.companion_tag.is_some()
+	}
+
 	/// Returns the [`TagType`]
 	pub fn tag_type(&self) -> TagType {
 		self.tag_type
@@ -341,12 +404,57 @@ impl Tag {
 	}
 
 	/// Removes all items with the specified [`ItemKey`], and returns them
+	///
+	/// See also: [take_filter()](Self::take_filter)
 	pub fn take(&mut self, key: &ItemKey) -> impl Iterator<Item = TagItem> + '_ {
+		self.take_filter(key, |_| true)
+	}
+
+	/// Removes selected items with the specified [`ItemKey`], and returns them
+	///
+	/// Only takes items for which `filter()` returns `true`. All other items are retained.
+	///
+	/// Returns the selected items in order and preserves the ordering of the remaining items.
+	///
+	/// # Examples
+	///
+	/// ```
+	/// use lofty::tag::{ItemKey, ItemValue, Tag, TagItem, TagType};
+	///
+	/// let mut tag = Tag::new(TagType::Id3v2);
+	/// tag.push(TagItem::new(
+	/// 	ItemKey::Comment,
+	/// 	ItemValue::Text("comment without description".to_owned()),
+	/// ));
+	/// let mut item = TagItem::new(
+	/// 	ItemKey::Comment,
+	/// 	ItemValue::Text("comment with description".to_owned()),
+	/// );
+	/// item.set_description("description".to_owned());
+	/// tag.push(item);
+	/// assert_eq!(tag.get_strings(&ItemKey::Comment).count(), 2);
+	///
+	/// // Extract all comment items with an empty description.
+	/// let comments = tag
+	/// 	.take_filter(&ItemKey::Comment, |item| item.description().is_empty())
+	/// 	.filter_map(|item| item.into_value().into_string())
+	/// 	.collect::<Vec<_>>();
+	/// assert_eq!(comments, vec!["comment without description".to_owned()]);
+	///
+	/// // The comments that didn't match the filter are still present.
+	/// assert_eq!(tag.get_strings(&ItemKey::Comment).count(), 1);
+	/// ```
+	pub fn take_filter(
+		&mut self,
+		key: &ItemKey,
+		mut filter: impl FnMut(&TagItem) -> bool,
+	) -> impl Iterator<Item = TagItem> + '_ {
 		// TODO: drain_filter
-		let mut split_idx = 0_usize;
+		let mut split_idx = 0;
 
 		for read_idx in 0..self.items.len() {
-			if self.items[read_idx].key() == key {
+			let item = &self.items[read_idx];
+			if item.key() == key && filter(item) {
 				self.items.swap(split_idx, read_idx);
 				split_idx += 1;
 			}
@@ -450,27 +558,38 @@ impl Tag {
 	/// # Examples
 	///
 	/// ```rust
-	/// use lofty::{Picture, Tag, TagType};
-	/// # use lofty::{PictureType, MimeType};
+	/// use lofty::picture::{MimeType, Picture, PictureType};
+	/// use lofty::tag::{Tag, TagType};
 	///
-	/// # let front_cover = Picture::new_unchecked(PictureType::CoverFront, Some(MimeType::Png), None, Vec::new());
-	/// # let back_cover = Picture::new_unchecked(PictureType::CoverBack, Some(MimeType::Png), None, Vec::new());
-	/// # let another_picture = Picture::new_unchecked(PictureType::Band, Some(MimeType::Png), None, Vec::new());
 	/// let mut tag = Tag::new(TagType::Id3v2);
 	///
 	/// // Add a front cover
+	/// let front_cover = Picture::new_unchecked(
+	/// 	PictureType::CoverFront,
+	/// 	Some(MimeType::Png),
+	/// 	None,
+	/// 	Vec::new(),
+	/// );
 	/// tag.push_picture(front_cover);
 	///
 	/// assert_eq!(tag.pictures().len(), 1);
 	/// assert_eq!(tag.pictures()[0].pic_type(), PictureType::CoverFront);
 	///
 	/// // Replace the front cover with a back cover
+	/// let back_cover = Picture::new_unchecked(
+	/// 	PictureType::CoverBack,
+	/// 	Some(MimeType::Png),
+	/// 	None,
+	/// 	Vec::new(),
+	/// );
 	/// tag.set_picture(0, back_cover);
 	///
 	/// assert_eq!(tag.pictures().len(), 1);
 	/// assert_eq!(tag.pictures()[0].pic_type(), PictureType::CoverBack);
 	///
 	/// // Use an out of bounds index
+	/// let another_picture =
+	/// 	Picture::new_unchecked(PictureType::Band, Some(MimeType::Png), None, Vec::new());
 	/// tag.set_picture(100, another_picture);
 	///
 	/// assert_eq!(tag.pictures().len(), 2);
@@ -492,11 +611,18 @@ impl Tag {
 	/// # Examples
 	///
 	/// ```rust
-	/// use lofty::{Picture, Tag, TagType};
-	/// # use lofty::{PictureType, MimeType};
+	/// use lofty::picture::{MimeType, Picture, PictureType};
+	/// use lofty::tag::{Tag, TagType};
 	///
-	/// # let picture = Picture::new_unchecked(PictureType::CoverFront, Some(MimeType::Png), None, Vec::new());
 	/// let mut tag = Tag::new(TagType::Id3v2);
+	///
+	/// let picture = Picture::new_unchecked(
+	/// 	PictureType::CoverFront,
+	/// 	Some(MimeType::Png),
+	/// 	None,
+	/// 	Vec::new(),
+	/// );
+	///
 	/// tag.push_picture(picture);
 	///
 	/// assert_eq!(tag.pictures().len(), 1);
@@ -514,6 +640,11 @@ impl TagExt for Tag {
 	type Err = LoftyError;
 	type RefKey<'a> = &'a ItemKey;
 
+	#[inline]
+	fn tag_type(&self) -> TagType {
+		self.tag_type
+	}
+
 	fn len(&self) -> usize {
 		self.items.len() + self.pictures.len()
 	}
@@ -526,19 +657,28 @@ impl TagExt for Tag {
 		self.items.is_empty() && self.pictures.is_empty()
 	}
 
-	/// Save the `Tag` to a [`File`](std::fs::File)
+	/// Save the `Tag` to a [`FileLike`]
 	///
 	/// # Errors
 	///
-	/// * A [`FileType`](crate::FileType) couldn't be determined from the File
-	/// * Attempting to write a tag to a format that does not support it. See [`FileType::supports_tag_type`](crate::FileType::supports_tag_type)
-	fn save_to(&self, file: &mut File) -> std::result::Result<(), Self::Err> {
+	/// * A [`FileType`](crate::file::FileType) couldn't be determined from the File
+	/// * Attempting to write a tag to a format that does not support it. See [`FileType::supports_tag_type`](crate::file::FileType::supports_tag_type)
+	fn save_to<F>(
+		&self,
+		file: &mut F,
+		write_options: WriteOptions,
+	) -> std::result::Result<(), Self::Err>
+	where
+		F: FileLike,
+		LoftyError: From<<F as Truncate>::Error>,
+		LoftyError: From<<F as Length>::Error>,
+	{
 		let probe = Probe::new(file).guess_file_type()?;
 
 		match probe.file_type() {
 			Some(file_type) => {
 				if file_type.supports_tag_type(self.tag_type()) {
-					utils::write_tag(self, probe.into_inner(), file_type)
+					utils::write_tag(self, probe.into_inner(), file_type, write_options)
 				} else {
 					err!(UnsupportedTag);
 				}
@@ -547,8 +687,8 @@ impl TagExt for Tag {
 		}
 	}
 
-	fn dump_to<W: Write>(&self, writer: &mut W) -> Result<()> {
-		utils::dump_tag(self, writer)
+	fn dump_to<W: Write>(&self, writer: &mut W, write_options: WriteOptions) -> Result<()> {
+		utils::dump_tag(self, writer, write_options)
 	}
 
 	/// Remove a tag from a [`Path`]
@@ -560,12 +700,17 @@ impl TagExt for Tag {
 		self.tag_type.remove_from_path(path)
 	}
 
-	/// Remove a tag from a [`File`]
+	/// Remove a tag from a [`FileLike`]
 	///
 	/// # Errors
 	///
 	/// See [`TagType::remove_from`]
-	fn remove_from(&self, file: &mut File) -> std::result::Result<(), Self::Err> {
+	fn remove_from<F>(&self, file: &mut F) -> std::result::Result<(), Self::Err>
+	where
+		F: FileLike,
+		LoftyError: From<<F as Truncate>::Error>,
+		LoftyError: From<<F as Length>::Error>,
+	{
 		self.tag_type.remove_from(file)
 	}
 
@@ -576,6 +721,7 @@ impl TagExt for Tag {
 }
 
 #[derive(Debug, Clone, Default)]
+#[allow(missing_docs)]
 pub struct SplitTagRemainder;
 
 impl SplitTag for Tag {
@@ -594,72 +740,15 @@ impl MergeTag for SplitTagRemainder {
 	}
 }
 
-/// The tag's format
-#[derive(Copy, Clone, Debug, PartialEq, Eq)]
-#[non_exhaustive]
-pub enum TagType {
-	/// This covers both APEv1 and APEv2 as it doesn't matter much
-	Ape,
-	/// Represents an ID3v1 tag
-	Id3v1,
-	/// This covers all ID3v2 versions since they all get upgraded to ID3v2.4
-	Id3v2,
-	/// Represents an MP4 ilst atom
-	Mp4Ilst,
-	/// Represents vorbis comments
-	VorbisComments,
-	/// Represents a RIFF INFO LIST
-	RiffInfo,
-	/// Represents AIFF text chunks
-	AiffText,
-}
-
-impl TagType {
-	/// Remove a tag from a [`Path`]
-	///
-	/// # Errors
-	///
-	/// See [`TagType::remove_from`]
-	pub fn remove_from_path(&self, path: impl AsRef<Path>) -> Result<()> {
-		let mut file = OpenOptions::new().read(true).write(true).open(path)?;
-		self.remove_from(&mut file)
-	}
-
-	#[allow(clippy::shadow_unrelated)]
-	/// Remove a tag from a [`File`]
-	///
-	/// # Errors
-	///
-	/// * It is unable to guess the file format
-	/// * The format doesn't support the tag
-	/// * It is unable to write to the file
-	pub fn remove_from(&self, file: &mut File) -> Result<()> {
-		let probe = Probe::new(file).guess_file_type()?;
-		let Some(file_type) = probe.file_type() else {
-			err!(UnknownFormat);
-		};
-
-		// TODO: This should not have to be manually updated
-		let special_exceptions = ((file_type == FileType::Ape
-			|| file_type == FileType::Mpc
-			|| file_type == FileType::Flac)
-			&& *self == TagType::Id3v2)
-			|| file_type == FileType::Mpc && *self == TagType::Id3v1;
-
-		if !special_exceptions && !file_type.supports_tag_type(*self) {
-			err!(UnsupportedTag);
-		}
-
-		let file = probe.into_inner();
-		utils::write_tag(&Tag::new(*self), file, file_type)
-	}
-}
-
 #[cfg(test)]
 mod tests {
 	use super::try_parse_year;
+	use crate::config::WriteOptions;
+	use crate::picture::{Picture, PictureType};
+	use crate::prelude::*;
 	use crate::tag::utils::test_utils::read_path;
-	use crate::{Accessor, Picture, PictureType, Tag, TagExt, TagType};
+	use crate::tag::{Tag, TagType};
+
 	use std::io::{Seek, Write};
 	use std::process::Command;
 
@@ -677,7 +766,8 @@ mod tests {
 		picture.set_pic_type(PictureType::CoverFront);
 
 		tag.push_picture(picture);
-		tag.save_to(temp_file.as_file_mut()).unwrap();
+		tag.save_to(temp_file.as_file_mut(), WriteOptions::default())
+			.unwrap();
 
 		let cmd_output = Command::new("ffprobe")
 			.arg(temp_file.path().to_str().unwrap())
@@ -719,7 +809,8 @@ mod tests {
 		picture.set_pic_type(PictureType::CoverFront);
 
 		tag.push_picture(picture);
-		tag.save_to(temp_file.as_file_mut()).unwrap();
+		tag.save_to(temp_file.as_file_mut(), WriteOptions::default())
+			.unwrap();
 
 		let cmd_output = Command::new("opusinfo")
 			.arg(temp_file.path().to_str().unwrap())
